@@ -9,7 +9,7 @@
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU Library General Public License for more details.
+ *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
@@ -225,150 +225,162 @@
 	save(op1,lf_resd);										\
 	lflags.type=t_DECd;
 
-#define ROLB(op1,op2,load,save)								\
-		LoadZF;LoadSF;LoadAF;								\
-		lf_var1b=load(op1);								\
-		lf_var2b=op2&0x07;								\
-		lf_resb=(lf_var1b << lf_var2b) |		\
-				(lf_var1b >> (8-lf_var2b));			\
-		save(op1,lf_resb);							\
-		lflags.type=t_ROLb;									\
 
 
-#define ROLW(op1,op2,load,save)								\
-		LoadZF;LoadSF;LoadAF;								\
-		lf_var1w=load(op1);								\
-		lf_var2b=op2&0x0F;								\
-		lf_resw=(lf_var1w << lf_var2b) |		\
-				(lf_var1w >> (16-lf_var2b));		\
-		save(op1,lf_resw);							\
-		lflags.type=t_ROLw;									\
+#define ROLB(op1,op2,load,save)						\
+	if (!(op2&0x7)) break;							\
+	FillFlags();									\
+	lf_var1b=load(op1);								\
+	lf_var2b=op2&0x07;								\
+	lf_resb=(lf_var1b << lf_var2b) |				\
+			(lf_var1b >> (8-lf_var2b));				\
+	save(op1,lf_resb);								\
+	SETFLAGBIT(CF,lf_resb & 1);						\
+	SETFLAGBIT(OF,(lf_resb & 1) ^ (lf_resb >> 7));
+
+#define ROLW(op1,op2,load,save)						\
+	if (!(op2&0xf)) break;						\
+	FillFlags();									\
+	lf_var1w=load(op1);								\
+	lf_var2b=op2&0xf;								\
+	lf_resw=(lf_var1w << lf_var2b) |				\
+			(lf_var1w >> (16-lf_var2b));			\
+	save(op1,lf_resw);								\
+	SETFLAGBIT(CF,lf_resw & 1);						\
+	SETFLAGBIT(OF,(lf_resw & 1) ^ (lf_resw >> 15));
+
+#define ROLD(op1,op2,load,save)						\
+	if (!op2) break;								\
+	FillFlags();									\
+	lf_var1d=load(op1);								\
+	lf_var2b=op2;									\
+	lf_resd=(lf_var1d << lf_var2b) |				\
+			(lf_var1d >> (32-lf_var2b));			\
+	save(op1,lf_resd);								\
+	SETFLAGBIT(CF,lf_resd & 1);						\
+	SETFLAGBIT(OF,(lf_resd & 1) ^ (lf_resd >> 31));
 
 
-#define ROLD(op1,op2,load,save)								\
-		LoadZF;LoadSF;LoadAF;								\
-		lf_var1d=load(op1);								\
-		lf_var2b=op2;									\
-		lf_resd=(lf_var1d << lf_var2b) |		\
-				(lf_var1d >> (32-lf_var2b));		\
-		save(op1,lf_resd);							\
-		lflags.type=t_ROLd;									\
+#define RORB(op1,op2,load,save)						\
+	if (!(op2&0x7)) break;							\
+	FillFlags();									\
+	lf_var1b=load(op1);								\
+	lf_var2b=op2&0x07;								\
+	lf_resb=(lf_var1b >> lf_var2b) |				\
+			(lf_var1b << (8-lf_var2b));				\
+	save(op1,lf_resb);								\
+	SETFLAGBIT(CF,lf_resb & 0x80);					\
+	if (lf_var2b == 1) SETFLAGBIT(OF,(lf_resb ^ lf_var1b) & 0x80);
+
+#define RORW(op1,op2,load,save)					\
+	if (!(op2&0xf)) break;						\
+	FillFlags();								\
+	lf_var1w=load(op1);							\
+	lf_var2b=op2&0xf;							\
+	lf_resw=(lf_var1w >> lf_var2b) |			\
+			(lf_var1w << (16-lf_var2b));		\
+	save(op1,lf_resw);							\
+	SETFLAGBIT(CF,lf_resw & 0x8000);			\
+	if (lf_var2b == 1) SETFLAGBIT(OF,(lf_resw ^ lf_var1w) & 0x8000);
+
+#define RORD(op1,op2,load,save)					\
+	if (!op2) break;							\
+	FillFlags();								\
+	lf_var1d=load(op1);							\
+	lf_var2b=op2;								\
+	lf_resd=(lf_var1d >> lf_var2b) |			\
+			(lf_var1d << (32-lf_var2b));		\
+	save(op1,lf_resd);							\
+	SETFLAGBIT(CF,lf_resd & 0x80000000);		\
+	if (lf_var2b == 1) SETFLAGBIT(OF,(lf_resd ^ lf_var1d) & 0x80000000);
 
 
-#define RORB(op1,op2,load,save)								\
-		LoadZF;LoadSF;LoadAF;								\
-		lf_var1b=load(op1);								\
-		lf_var2b=op2&0x07;								\
-		lf_resb=(lf_var1b >> lf_var2b) |		\
-				(lf_var1b << (8-lf_var2b));			\
-		save(op1,lf_resb);							\
-		lflags.type=t_RORb;									\
-
-
-#define RORW(op1,op2,load,save)								\
-	if (op2&0x0F) {											\
-		LoadZF;LoadSF;LoadAF;								\
-		lf_var1w=load(op1);								\
-		lf_var2b=op2&0x0F;								\
-		lf_resw=(lf_var1w >> lf_var2b) |		\
-				(lf_var1w << (16-lf_var2b));		\
-		save(op1,lf_resw);							\
-		lflags.type=t_RORw;									\
-	}
-
-#define RORD(op1,op2,load,save)								\
-	if (op2) {												\
-		LoadZF;LoadSF;LoadAF;								\
-		lf_var1d=load(op1);								\
-		lf_var2b=op2;									\
-		lf_resd=(lf_var1d >> lf_var2b) |		\
-				(lf_var1d << (32-lf_var2b));		\
-		save(op1,lf_resd);							\
-		lflags.type=t_RORd;									\
-	}
-
-
-#define RCLB(op1,op2,load,save)								\
-	if (op2%9) {											\
-		Bit8u cf=(Bit8u)FillFlags()&0x1;					\
-		lf_var1b=load(op1);									\
-		lf_var2b=op2%9;										\
-		lflags.type=t_RCLb;									\
-		lf_resb=(lf_var1b << lf_var2b) |					\
-				(cf << (lf_var2b-1)) |						\
-				(lf_var1b >> (9-lf_var2b));					\
-		SETFLAGBIT(CF,((lf_var1b >> (8-lf_var2b)) & 1));	\
- 		save(op1,lf_resb);									\
-	}
-
-#define RCLW(op1,op2,load,save)								\
-	if (op2%17) {											\
-		Bit16u cf=(Bit16u)FillFlags()&0x1;					\
-		lf_var1w=load(op1);								\
-		lf_var2b=op2%17;								\
-		lflags.type=t_RCLw;									\
-		lf_resw=(lf_var1w << lf_var2b) |		\
-				(cf << (lf_var2b-1)) |					\
-				(lf_var1w >> (17-lf_var2b));		\
-		SETFLAGBIT(CF,((lf_var1w >> (16-lf_var2b)) & 1));				\
-		save(op1,lf_resw);							\
-	}
-
-#define RCLD(op1,op2,load,save)								\
-	if (op2) {												\
-		Bit32u cf=(Bit32u)FillFlags()&0x1;					\
-		lf_var1d=load(op1);								\
-		lf_var2b=op2;									\
-		lflags.type=t_RCLd;									\
-		if (lf_var2b==1)	{							\
-			lf_resd=(lf_var1d << 1) | cf;		\
-		} else 	{											\
-			lf_resd=(lf_var1d << lf_var2b) |	\
+#define RCLB(op1,op2,load,save)							\
+	if (!(op2%9)) break;								\
+{	Bit8u cf=(Bit8u)FillFlags()&0x1;					\
+	lf_var1b=load(op1);									\
+	lf_var2b=op2%9;										\
+	lf_resb=(lf_var1b << lf_var2b) |					\
 			(cf << (lf_var2b-1)) |						\
-			(lf_var1d >> (33-lf_var2b));			\
-		}													\
-		SETFLAGBIT(CF,((lf_var1d >> (32-lf_var2b)) & 1));				\
-		save(op1,lf_resd);							\
-	}
+			(lf_var1b >> (9-lf_var2b));					\
+ 	save(op1,lf_resb);									\
+	SETFLAGBIT(CF,((lf_var1b >> (8-lf_var2b)) & 1));	\
+	SETFLAGBIT(OF,(reg_flags & 1) ^ (lf_resb >> 7));	\
+}
+
+#define RCLW(op1,op2,load,save)							\
+	if (!(op2%17)) break;								\
+{	Bit16u cf=(Bit16u)FillFlags()&0x1;					\
+	lf_var1w=load(op1);									\
+	lf_var2b=op2%17;									\
+	lf_resw=(lf_var1w << lf_var2b) |					\
+			(cf << (lf_var2b-1)) |						\
+			(lf_var1w >> (17-lf_var2b));				\
+	save(op1,lf_resw);									\
+	SETFLAGBIT(CF,((lf_var1w >> (16-lf_var2b)) & 1));	\
+	SETFLAGBIT(OF,(reg_flags & 1) ^ (lf_resw >> 15));	\
+}
+
+#define RCLD(op1,op2,load,save)							\
+	if (!op2) break;									\
+{	Bit32u cf=(Bit32u)FillFlags()&0x1;					\
+	lf_var1d=load(op1);									\
+	lf_var2b=op2;										\
+	if (lf_var2b==1)	{								\
+		lf_resd=(lf_var1d << 1) | cf;					\
+	} else 	{											\
+		lf_resd=(lf_var1d << lf_var2b) |				\
+		(cf << (lf_var2b-1)) |							\
+		(lf_var1d >> (33-lf_var2b));					\
+	}													\
+	save(op1,lf_resd);									\
+	SETFLAGBIT(CF,((lf_var1d >> (32-lf_var2b)) & 1));	\
+	SETFLAGBIT(OF,(reg_flags & 1) ^ (lf_resd >> 31));	\
+}
+
+
 
 #define RCRB(op1,op2,load,save)								\
 	if (op2%9) {											\
 		Bit8u cf=(Bit8u)FillFlags()&0x1;					\
-		lf_var1b=load(op1);								\
-		lf_var2b=op2%9;									\
-		lflags.type=t_RCRb;									\
-	 	lf_resb=(lf_var1b >> lf_var2b) |		\
-				(cf << (8-lf_var2b)) |					\
-				(lf_var1b << (9-lf_var2b));			\
-		save(op1,lf_resb);							\
+		lf_var1b=load(op1);									\
+		lf_var2b=op2%9;										\
+	 	lf_resb=(lf_var1b >> lf_var2b) |					\
+				(cf << (8-lf_var2b)) |						\
+				(lf_var1b << (9-lf_var2b));					\
+		save(op1,lf_resb);									\
+		SETFLAGBIT(CF,(lf_var1b >> (lf_var2b - 1)) & 1);	\
+		SETFLAGBIT(OF,(lf_resb ^ lf_var1b) & 0x80);			\
 	}
 
 #define RCRW(op1,op2,load,save)								\
 	if (op2%17) {											\
 		Bit16u cf=(Bit16u)FillFlags()&0x1;					\
-		lf_var1w=load(op1);								\
-		lf_var2b=op2%17;								\
-		lflags.type=t_RCRw;									\
-	 	lf_resw=(lf_var1w >> lf_var2b) |		\
-				(cf << (16-lf_var2b)) |					\
-				(lf_var1w << (17-lf_var2b));		\
-		save(op1,lf_resw);							\
+		lf_var1w=load(op1);									\
+		lf_var2b=op2%17;									\
+	 	lf_resw=(lf_var1w >> lf_var2b) |					\
+				(cf << (16-lf_var2b)) |						\
+				(lf_var1w << (17-lf_var2b));				\
+		save(op1,lf_resw);									\
+		SETFLAGBIT(CF,(lf_var1w >> (lf_var2b - 1)) & 1);	\
+		SETFLAGBIT(OF,(lf_resw ^ lf_var1w) & 0x8000);		\
 	}
 
 #define RCRD(op1,op2,load,save)								\
 	if (op2) {												\
 		Bit32u cf=(Bit32u)FillFlags()&0x1;					\
-		lf_var1d=load(op1);								\
-		lf_var2b=op2;									\
-		lflags.type=t_RCRd;									\
-		if (lf_var2b==1) {								\
-			lf_resd=lf_var1d >> 1 | cf << 31;	\
+		lf_var1d=load(op1);									\
+		lf_var2b=op2;										\
+		if (lf_var2b==1) {									\
+			lf_resd=lf_var1d >> 1 | cf << 31;				\
 		} else {											\
- 			lf_resd=(lf_var1d >> lf_var2b) |	\
-				(cf << (32-lf_var2b)) |					\
-				(lf_var1d << (33-lf_var2b));		\
+ 			lf_resd=(lf_var1d >> lf_var2b) |				\
+				(cf << (32-lf_var2b)) |						\
+				(lf_var1d << (33-lf_var2b));				\
 		}													\
-		save(op1,lf_resd);							\
+		save(op1,lf_resd);									\
+		SETFLAGBIT(CF,(lf_var1d >> (lf_var2b - 1)) & 1);	\
+		SETFLAGBIT(OF,(lf_resd ^ lf_var1d) & 0x80000000);	\
 	}
 
 
@@ -548,7 +560,7 @@
 	}
 
 #define MULB(op1,load,save)									\
-	lflags.type=t_MUL;										\
+	FillFlags();											\
 	reg_ax=reg_al*load(op1);								\
 	if (reg_ax & 0xff00) {									\
 		SETFLAGBIT(CF,true);SETFLAGBIT(OF,true);			\
@@ -558,10 +570,10 @@
 
 #define MULW(op1,load,save)									\
 {															\
+	FillFlags();											\
 	Bitu tempu=(Bitu)reg_ax*(Bitu)(load(op1));				\
 	reg_ax=(Bit16u)(tempu);									\
 	reg_dx=(Bit16u)(tempu >> 16);							\
-	lflags.type=t_MUL;										\
 	if (reg_dx) {											\
 		SETFLAGBIT(CF,true);SETFLAGBIT(OF,true);			\
 	} else {												\
@@ -571,10 +583,10 @@
 
 #define MULD(op1,load,save)									\
 {															\
+	FillFlags();											\
 	Bit64u tempu=(Bit64u)reg_eax*(Bit64u)(load(op1));		\
 	reg_eax=(Bit32u)(tempu);								\
 	reg_edx=(Bit32u)(tempu >> 32);							\
-	lflags.type=t_MUL;										\
 	if (reg_edx) {											\
 		SETFLAGBIT(CF,true);SETFLAGBIT(OF,true);			\
 	} else {												\
@@ -651,7 +663,7 @@
 
 #define IMULB(op1,load,save)								\
 {															\
-	lflags.type=t_MUL;										\
+	FillFlags();											\
 	reg_ax=((Bit8s)reg_al) * ((Bit8s)(load(op1)));			\
 	if ((reg_ax & 0xff80)==0xff80 ||						\
 		(reg_ax & 0xff80)==0x0000) {						\
@@ -664,10 +676,10 @@
 
 #define IMULW(op1,load,save)								\
 {															\
+	FillFlags();											\
 	Bits temps=((Bit16s)reg_ax)*((Bit16s)(load(op1)));		\
 	reg_ax=(Bit16s)(temps);									\
 	reg_dx=(Bit16s)(temps >> 16);							\
-	lflags.type=t_MUL;										\
 	if (((temps & 0xffff8000)==0xffff8000 ||				\
 		(temps & 0xffff8000)==0x0000)) {					\
 		SETFLAGBIT(CF,false);SETFLAGBIT(OF,false);			\
@@ -678,11 +690,11 @@
 
 #define IMULD(op1,load,save)								\
 {															\
+	FillFlags();											\
 	Bit64s temps=((Bit64s)((Bit32s)reg_eax))*				\
 				 ((Bit64s)((Bit32s)(load(op1))));			\
 	reg_eax=(Bit32u)(temps);								\
 	reg_edx=(Bit32u)(temps >> 32);							\
-	lflags.type=t_MUL;										\
 	if ((reg_edx==0xffffffff) &&							\
 		(reg_eax & 0x80000000) ) {							\
 		SETFLAGBIT(CF,false);SETFLAGBIT(OF,false);			\
@@ -696,10 +708,10 @@
 
 #define DIMULW(op1,op2,op3,load,save)						\
 {															\
+	FillFlags();											\
 	Bits res;												\
 	res=((Bit16s)op2) * ((Bit16s)op3);						\
 	save(op1,res & 0xffff);									\
-	lflags.type=t_MUL;										\
 	if ((res> -32768)  && (res<32767)) {					\
 		SETFLAGBIT(CF,false);SETFLAGBIT(OF,false);			\
 	} else {												\
@@ -709,9 +721,9 @@
 
 #define DIMULD(op1,op2,op3,load,save)						\
 {															\
+	FillFlags();											\
 	Bit64s res=((Bit64s)((Bit32s)op2))*((Bit64s)((Bit32s)op3));	\
 	save(op1,(Bit32s)res);									\
-	lflags.type=t_MUL;										\
 	if ((res>-((Bit64s)(2147483647)+1)) &&					\
 		(res<(Bit64s)2147483647)) {							\
 		SETFLAGBIT(CF,false);SETFLAGBIT(OF,false);			\
