@@ -9,14 +9,14 @@
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU Library General Public License for more details.
+ *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-/* $Id: dos.cpp,v 1.65 2004/02/02 11:38:44 qbix79 Exp $ */
+/* $Id: dos.cpp,v 1.77 2004/11/16 14:24:52 qbix79 Exp $ */
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -36,7 +36,6 @@ DOS_InfoBlock dos_infoblock;
 
 Bit8u dos_copybuf[0x10000];
 static Bitu call_20,call_21,call_25,call_26,call_27,call_28,call_29;
-static Bitu call_casemap;
 
 void DOS_SetError(Bit16u code) {
 	dos.errorcode=code;
@@ -45,20 +44,20 @@ void DOS_SetError(Bit16u code) {
 #define DOSNAMEBUF 256
 static Bitu DOS_21Handler(void) {
 	if (((reg_ah != 0x50) && (reg_ah != 0x51) && (reg_ah != 0x62) && (reg_ah != 0x64)) && (reg_ah<0x6c)) {
-		DOS_PSP psp(dos.psp);
-		psp.SetStack(RealMake(SegValue(ss),reg_sp-20));
+		DOS_PSP psp(dos.psp());
+		psp.SetStack(RealMake(SegValue(ss),reg_sp-18));
 	}
+
 	char name1[DOSNAMEBUF+1];
 	char name2[DOSNAMEBUF+1];
 	switch (reg_ah) {
 	case 0x01:		/* Read character from STDIN, with echo */
 		{	
 			Bit8u c;Bit16u n=1;
-             dos.echo=true;
+			dos.echo=true;
 			DOS_ReadFile(STDIN,&c,&n);
 			reg_al=c;
-            dos.echo=false;
-
+			dos.echo=false;
 		}
 		break;
 	case 0x02:		/* Write character to STDOUT */
@@ -181,31 +180,26 @@ static Bitu DOS_21Handler(void) {
 		break;
 	case 0x10:		/* Close File using FCB */
 		if(DOS_FCBClose(SegValue(ds),reg_dx)){
-		reg_al=0;
+			reg_al=0;
 		}else{
 			reg_al=0xff;
 		}
 		LOG(LOG_FCB,LOG_NORMAL)("DOS:0x10 FCB-fileclose used, result:al=%d",reg_al);
 		break;
 	case 0x11:		/* Find First Matching File using FCB */
-		if(DOS_FCBFindFirst(SegValue(ds),reg_dx)){
-		reg_al=0;
-		}else{
-			reg_al=0xff;
-		}
+		if(DOS_FCBFindFirst(SegValue(ds),reg_dx)) reg_al = 0x00;
+		else reg_al = 0xFF;
 		LOG(LOG_FCB,LOG_NORMAL)("DOS:0x11 FCB-FindFirst used, result:al=%d",reg_al);
 		break;
 	case 0x12:		/* Find Next Matching File using FCB */
-		if(DOS_FCBFindNext(SegValue(ds),reg_dx)){
-		reg_al=0;
-		}else{
-			reg_al=0xff;
-		}
+		if(DOS_FCBFindNext(SegValue(ds),reg_dx)) reg_al = 0x00;
+		else reg_al = 0xFF;
 		LOG(LOG_FCB,LOG_NORMAL)("DOS:0x12 FCB-FindNext used, result:al=%d",reg_al);
 		break;
 	case 0x13:		/* Delete File using FCB */
 		if (DOS_FCBDeleteFile(SegValue(ds),reg_dx)) reg_al = 0x00;
 		else reg_al = 0xFF;
+		LOG(LOG_FCB,LOG_NORMAL)("DOS:0x16 FCB-Delete used, result:al=%d",reg_al);
 		break;
 	case 0x14:		/* Sequential read from FCB */
 		reg_al = DOS_FCBRead(SegValue(ds),reg_dx,0);
@@ -254,29 +248,26 @@ static Bitu DOS_21Handler(void) {
 		LOG(LOG_FCB,LOG_NORMAL)("DOS:0x28 FCB-Random(block) write used, result:al=%d",reg_al);
 		break;
 	case 0x29:		/* Parse filename into FCB */
-        {   Bit8u difference;
-            char string[1024];
-            MEM_StrCopy(SegPhys(ds)+reg_si,string,1024);
-            reg_al=FCB_Parsename(SegValue(es),reg_di,reg_al ,string, &difference);
-            reg_si+=difference;
-        }
+		{   
+			Bit8u difference;
+			char string[1024];
+			MEM_StrCopy(SegPhys(ds)+reg_si,string,1024);
+			reg_al=FCB_Parsename(SegValue(es),reg_di,reg_al ,string, &difference);
+			reg_si+=difference;
+		}
 		LOG(LOG_FCB,LOG_NORMAL)("DOS:29:FCB Parse Filename, result:al=%d",reg_al);
-        break;
+		break;
 	case 0x19:		/* Get current default drive */
 		reg_al=DOS_GetDefaultDrive();
 		break;
 	case 0x1a:		/* Set Disk Transfer Area Address */
-		{
-			dos.dta=RealMakeSeg(ds,reg_dx);
-			DOS_PSP psp(dos.psp);
-			psp.SetDTA(dos.dta);
-		}
+		dos.dta(RealMakeSeg(ds,reg_dx));
 		break;
 	case 0x25:		/* Set Interrupt Vector */
 		RealSetVec(reg_al,RealMakeSeg(ds,reg_dx));
 		break;
 	case 0x26:		/* Create new PSP */
-		DOS_NewPSP(reg_dx,DOS_PSP(dos.psp).GetSize());
+		DOS_NewPSP(reg_dx,DOS_PSP(dos.psp()).GetSize());
 		break;
 	case 0x2a:		/* Get System Date */
 		{
@@ -319,29 +310,43 @@ static Bitu DOS_21Handler(void) {
 		dos.verify=(reg_al==1);
 		break;
 	case 0x2f:		/* Get Disk Transfer Area */
-		SegSet16(es,RealSeg(dos.dta));
-		reg_bx=RealOff(dos.dta);
+		SegSet16(es,RealSeg(dos.dta()));
+		reg_bx=RealOff(dos.dta());
 		break;
 	case 0x30:		/* Get DOS Version */
 		if (reg_al==0) reg_bh=0xFF;		/* Fake Microsoft DOS */
 		if (reg_al==1) reg_bh=0x10;		/* DOS is in HMA */
 		reg_al=dos.version.major;
 		reg_ah=dos.version.minor;
+		/* Serialnumber */
+		reg_bl=0x00;
+		reg_cx=0x0000;
 		break;
 	case 0x31:		/* Terminate and stay resident */
 		//TODO First get normal files executing
 		// Important: This service does not set the carry flag!
-		DOS_ResizeMemory(dos.psp,&reg_dx);
+		DOS_ResizeMemory(dos.psp(),&reg_dx);
 		DOS_Terminate(true);
-		dos.return_code=reg_al;
+		dos.return_code=reg_al; //Officially a field in the SDA
 		dos.return_mode=RETURN_TSR;
 		break;
+        case 0x32: /* Get drive parameter block for specific drive */
+		{	/* Officially a dpb should be returned as well. The disk detection part is implemented */
+			Bitu drive=reg_dl;if(!drive) drive=dos.current_drive;else drive--;
+			if(Drives[drive]) {
+				reg_al=0x00;
+				LOG(LOG_DOSMISC,LOG_ERROR)("Get drive parameter block.");   
+			} else {
+				reg_al=0xff;
+			}
+		}
+		break;
 	case 0x33:		/* Extended Break Checking */
 		switch (reg_al) {
 			case 0:reg_dl=dos.breakcheck;break;			/* Get the breakcheck flag */
 			case 1:dos.breakcheck=(reg_dl>0);break;		/* Set the breakcheck flag */
 			case 2:{bool old=dos.breakcheck;dos.breakcheck=(reg_dl>0);reg_dl=old;}break;
-			case 5:reg_dl=3;break;							/* Always boot from c: :) */
+			case 5:reg_dl=3;break;//TODO should be z						/* Always boot from c: :) */
 			case 6:											/* Get true version number */
 				reg_bl=dos.version.major;
 				reg_bh=dos.version.minor;
@@ -353,8 +358,8 @@ static Bitu DOS_21Handler(void) {
 		}
 		break;
 	case 0x34:		/* Get INDos Flag */
-		SegSet16(es,RealSeg(dos.tables.indosflag));
-		reg_bx=RealOff(dos.tables.indosflag);
+		SegSet16(es,DOS_SDA_SEG);
+		reg_bx=DOS_SDA_OFS + 0x01;
 		break;
 	case 0x35:		/* Get interrupt vector */
 		reg_bx=real_readw(0,((Bit16u)reg_al)*4);
@@ -363,8 +368,8 @@ static Bitu DOS_21Handler(void) {
 	case 0x36:		/* Get Free Disk Space */
 		{
 			Bit16u bytes,clusters,free;
-            Bit8u sectors;
-			if	(DOS_GetFreeDiskSpace(reg_dl,&bytes,&sectors,&clusters,&free)) {
+			Bit8u sectors;
+			if(DOS_GetFreeDiskSpace(reg_dl,&bytes,&sectors,&clusters,&free)) {
 				reg_ax=sectors;
 				reg_bx=free;
 				reg_cx=bytes;
@@ -390,18 +395,8 @@ static Bitu DOS_21Handler(void) {
 		break;
 	case 0x38:					/* Set Country Code */	
 		if (reg_al==0) {		/* Get country specidic information */
-			PhysPt pt = SegPhys(ds)+reg_dx;
-			mem_writew(pt   ,0x00); // USA
-			mem_writeb(pt+ 2, '$'); mem_writeb(pt+ 3,0x00);
-			mem_writeb(pt+ 7, '.'); mem_writeb(pt+ 8,0x00);
-			mem_writeb(pt+ 9, '.'); mem_writeb(pt+10,0x00);
-			mem_writeb(pt+11, '.'); mem_writeb(pt+12,0x00);
-			mem_writeb(pt+13, '.'); mem_writeb(pt+14,0x00);
-			mem_writeb(pt+15,0x01); // currency format
-			mem_writeb(pt+16,0x02);	// num digits
-			mem_writeb(pt+17,0x00); // time format
-			mem_writed(pt+18,CALLBACK_RealPointer(call_casemap));
-			mem_writew(pt+22,0x00); // data list seperator
+			PhysPt dest = SegPhys(ds)+reg_dx;
+			MEM_BlockWrite(dest,dos.tables.country,0x22);
 			reg_bx = 0x01;
 			CALLBACK_SCF(false);
 			break;
@@ -426,6 +421,7 @@ static Bitu DOS_21Handler(void) {
 		} else {
 			reg_ax=dos.errorcode;
 			CALLBACK_SCF(true);
+			LOG(LOG_MISC,LOG_NORMAL)("Remove dir failed on %s with error %X",name1,dos.errorcode);
 		}
 		break;
 	case 0x3b:		/* CHDIR Set current directory */
@@ -466,7 +462,7 @@ static Bitu DOS_21Handler(void) {
 	case 0x3f:		/* READ Read from file or device */
 		{ 
 			Bit16u toread=reg_cx;
-            dos.echo=true;
+			dos.echo=true;
 			if (DOS_ReadFile(reg_bx,dos_copybuf,&toread)) {
 				MEM_BlockWrite(SegPhys(ds)+reg_dx,dos_copybuf,toread);
 				reg_ax=toread;
@@ -475,7 +471,7 @@ static Bitu DOS_21Handler(void) {
 				reg_ax=dos.errorcode;
 				CALLBACK_SCF(true);
 			}
-            dos.echo=false;
+			dos.echo=false;
 			break;
 		}
 	case 0x40:					/* WRITE Write to file or device */
@@ -621,10 +617,9 @@ static Bitu DOS_21Handler(void) {
 		break;
 //TODO Check for use of execution state AL=5
 	case 0x00:
-           reg_ax=0x4c00;       /* Terminate Program */
+		reg_ax=0x4c00;				/* Terminate Program */
 	case 0x4c:					/* EXIT Terminate with return code */
-		
-        {
+	        {
 			if (DOS_Terminate(false)) {
 				/* This can't ever return false normally */
 			} else {            
@@ -634,7 +629,7 @@ static Bitu DOS_21Handler(void) {
 			break;
 		}
 	case 0x4d:					/* Get Return code */
-		reg_al=dos.return_code;
+		reg_al=dos.return_code;/* Officially read from SDA and clear when read */
 		reg_ah=dos.return_mode;
 		break;
 	case 0x4e:					/* FINDFIRST Find first matching file */
@@ -650,17 +645,18 @@ static Bitu DOS_21Handler(void) {
 	case 0x4f:					/* FINDNEXT Find next matching file */
 		if (DOS_FindNext()) {
 			CALLBACK_SCF(false);
-			reg_ax=0xffff;			/* Undocumented */
+			/* reg_ax=0xffff;*/			/* Undocumented */
+			reg_ax=0;				/* Undocumented:Qbix Willy beamish */
 		} else {
 			reg_ax=dos.errorcode;
 			CALLBACK_SCF(true);
 		};
 		break;		
 	case 0x50:					/* Set current PSP */
-		dos.psp=reg_bx;
+		dos.psp(reg_bx);
 		break;
 	case 0x51:					/* Get current PSP */
-		reg_bx=dos.psp;
+		reg_bx=dos.psp();
 		break;
 	case 0x52: {				/* Get list of lists */
 		RealPt addr=dos_infoblock.GetPointer();
@@ -671,14 +667,14 @@ static Bitu DOS_21Handler(void) {
 //TODO Think hard how shit this is gonna be
 //And will any game ever use this :)
 	case 0x53:					/* Translate BIOS parameter block to drive parameter block */
-        E_Exit("Unhandled Dos 21 call %02X",reg_ah);
-        break;
+		E_Exit("Unhandled Dos 21 call %02X",reg_ah);
+		break;
 	case 0x54:					/* Get verify flag */
-        reg_al=dos.verify?1:0;
+		reg_al=dos.verify?1:0;
 		break;
 	case 0x55:					/* Create Child PSP*/
 		DOS_ChildPSP(reg_dx,reg_si);
-		dos.psp = reg_dx;
+		dos.psp(reg_dx);
 		break;
 	case 0x56:					/* RENAME Rename file */
 		MEM_StrCopy(SegPhys(ds)+reg_dx,name1,DOSNAMEBUF);
@@ -736,6 +732,7 @@ static Bitu DOS_21Handler(void) {
 			MEM_StrCopy(SegPhys(ds)+reg_dx,name1,DOSNAMEBUF);
 			if (DOS_CreateTempFile(name1,&handle)) {
 				reg_ax=handle;
+				MEM_BlockWrite(SegPhys(ds)+reg_dx,name1,strlen(name1)+1);
 				CALLBACK_SCF(false);
 			} else {
 				reg_ax=dos.errorcode;
@@ -763,6 +760,19 @@ static Bitu DOS_21Handler(void) {
 			}
 			break;
 		}
+	case 0x5d:					/* Network Functions */
+		if(reg_al == 0x06) {
+			SegSet16(ds,DOS_SDA_SEG);
+			reg_si = DOS_SDA_OFS;
+			reg_cx = 0x80;  // swap if in dos
+			reg_dx = 0x1a;  // swap always
+			LOG(LOG_DOSMISC,LOG_ERROR)("Get SDA, Let's hope for the best!");
+		}
+		break;
+	case 0x5f:					/* Network redirection */
+		reg_ax=0x0001;		//Failing it
+		CALLBACK_SCF(true);
+		break; 
 	case 0x60:					/* Canonicalize filename or path */
 		MEM_StrCopy(SegPhys(ds)+reg_si,name1,DOSNAMEBUF);
 		if (DOS_Canonicalize(name1,name2)) {
@@ -774,23 +784,39 @@ static Bitu DOS_21Handler(void) {
 			}
 			break;
 	case 0x62:					/* Get Current PSP Address */
-		reg_bx=dos.psp;
+		reg_bx=dos.psp();
+		break;
+	case 0x63:					/* DOUBLE BYTE CHARACTER SET */
+		if(reg_al == 0) {
+			SegSet16(ds,RealSeg(dos.tables.dcbs));
+			reg_si=RealOff(dos.tables.dcbs);		
+			reg_al = 0;
+			CALLBACK_SCF(false); //undocumented
+		} else reg_al = 0xff; //Doesn't officially touch carry flag
 		break;
 	case 0x64:					/* Set device driver lookahead flag */
 		E_Exit("Unhandled Dos 21 call %02X",reg_ah);
 		break;
 	case 0x65:					/* Get extented country information and a lot of other useless shit*/
-		/* Todo maybe fully support this for now we set it standard for USA */
-		{
-			LOG(LOG_DOSMISC,LOG_ERROR)("DOS:65:Extended country information call");
+		{ /* Todo maybe fully support this for now we set it standard for USA */ 
+			LOG(LOG_DOSMISC,LOG_ERROR)("DOS:65:Extended country information call %X",reg_ax);
+			if(reg_cx < 0x05 ) {
+				DOS_SetError(DOSERR_FUNCTION_NUMBER_INVALID);
+				CALLBACK_SCF(true);
+				break;
+			}
 			PhysPt data=SegPhys(es)+reg_di;
 			switch (reg_al) {
 			case 1:
-				mem_writeb(data,reg_al);
-				mem_writew(data+1,4);
-				mem_writew(data+3,1);				
-				mem_writew(data+5,37);
-				reg_cx=4;
+				mem_writeb(data + 0x00,reg_al);
+				mem_writew(data + 0x01,0x26);
+				mem_writew(data + 0x03,1);
+				if(reg_cx > 0x06 ) mem_writew(data+0x05,0x01b5);
+				if(reg_cx > 0x08 ) {
+					Bitu amount = (reg_cx>=0x29)?0x22:(reg_cx-7);
+					MEM_BlockWrite(data + 0x07,dos.tables.country,amount);
+					reg_cx=(reg_cx>=0x29)?0x29:reg_cx;
+				}
 				CALLBACK_SCF(false);
 				break;
 			case 2:	// Get pointer to uppercase table
@@ -799,9 +825,9 @@ static Bitu DOS_21Handler(void) {
 			case 5: // Get pointer to filename terminator table
 			case 6: // Get pointer to collating sequence table
 			case 7: // Get pointer to double byte char set table
-				mem_writew(data  ,0x0000);	// We dont have this table...
-				mem_writew(data+2,0x0000);	// End of table
-				reg_cx=4;
+				mem_writeb(data,reg_al);
+				mem_writed(data+1,dos.tables.dcbs); //used to be 0
+				reg_cx=5;
 				CALLBACK_SCF(false);
 				break;
 			default:
@@ -821,7 +847,7 @@ static Bitu DOS_21Handler(void) {
 	case 0x67:					/* Set handle count */
 		/* Weird call to increase amount of file handles needs to allocate memory if >20 */
 		{
-			DOS_PSP psp(dos.psp);
+			DOS_PSP psp(dos.psp());
 			psp.SetNumFiles(reg_bx);
 			CALLBACK_SCF(false);
 			break;
@@ -854,8 +880,8 @@ static Bitu DOS_21Handler(void) {
 		CALLBACK_SCF(true);
 		LOG(LOG_DOSMISC,LOG_NORMAL)("DOS:Windows long file name support call %2X",reg_al);
 		break;
+
     case 0x68:                  /* FFLUSH Commit file */
-    case 0x63:					/* Weirdo double byte stuff (fails but say it succeeded) available only in MSDOS 2.25  */
         CALLBACK_SCF(false);    //mirek
     case 0xE0:
     case 0x18:	            	/* NULL Function for CP/M compatibility or Extended rename FCB */
@@ -865,13 +891,10 @@ static Bitu DOS_21Handler(void) {
     case 0x6b:		            /* NULL Function */
     case 0x61:		            /* UNUSED */
     case 0xEF:                  /* Used in Ancient Art Of War CGA */
-	case 0x5d:					/* Network Functions ||HMMM seems to critical error info and return 1!! Maybe implement it.??*/
-	                            /* al=06 clears cf and leaves al=6 and returns crit error flag location*/
 	case 0x1f:					/* Get drive parameter block for default drive */
-	case 0x32:					/* Get drive parameter block for specific drive */
+
 	case 0x5c:					/* FLOCK File region locking */
 	case 0x5e:					/* More Network Functions */
-	case 0x5f:					/* And Even More Network Functions */
     default:
         LOG(LOG_DOSMISC,LOG_ERROR)("DOS:Unhandled call %02X al=%02X. Set al to default of 0",reg_ah,reg_al);
         reg_al=0x00; /* default value */
@@ -894,7 +917,7 @@ static Bitu DOS_27Handler(void)
 {
 	// Terminate & stay resident
 	Bit16u para = (reg_dx/16)+((reg_dx % 16)>0);
-	if (DOS_ResizeMemory(dos.psp,&para)) DOS_Terminate(true);
+	if (DOS_ResizeMemory(dos.psp(),&para)) DOS_Terminate(true);
 	return CBRET_NONE;
 }
 static Bitu DOS_25Handler(void) {
@@ -903,9 +926,10 @@ static Bitu DOS_25Handler(void) {
 		SETFLAGBIT(CF,true);
 	}else{
 		SETFLAGBIT(CF,false);
-		reg_ax=0;
 		if((reg_cx != 1) ||(reg_dx != 1))
-			LOG(LOG_DOSMISC,LOG_NORMAL)("int 25 called but not as diskdetection");
+			LOG(LOG_DOSMISC,LOG_NORMAL)("int 25 called but not as diskdetection drive %X",reg_al);
+
+	   reg_ax=0;
 	}
     return CBRET_NONE;
 }
@@ -925,14 +949,16 @@ static Bitu DOS_28Handler(void) {
 }
 
 static Bitu DOS_29Handler(void) {
-    LOG(LOG_DOSMISC,LOG_ERROR)("int 29 called");
-    return CBRET_NONE;
+	static bool int29warn=false;
+	if(!int29warn) { 
+		LOG(LOG_DOSMISC,LOG_WARN)("Int 29 called. Redirecting to int 10:0x0e");
+		int29warn=true;
+	}
+	reg_ah=0x0e;
+	CALLBACK_RunRealInt(0x10);
+	return CBRET_NONE;
 }
 
-static Bitu DOS_CaseMapFunc(void) {
-    //LOG(LOG_DOSMISC,LOG_ERROR)("Case map routine called : %c",reg_al);
-    return CBRET_NONE;
-};
 
 void DOS_ShutDown(Section* sec)
 {	
@@ -942,32 +968,32 @@ void DOS_ShutDown(Section* sec)
 
 void DOS_Init(Section* sec) {
 	call_20=CALLBACK_Allocate();
-	CALLBACK_Setup(call_20,DOS_20Handler,CB_IRET);
+	CALLBACK_Setup(call_20,DOS_20Handler,CB_IRET,"DOS Int 20");
 	RealSetVec(0x20,CALLBACK_RealPointer(call_20));
 
 	call_21=CALLBACK_Allocate();
-	CALLBACK_Setup(call_21,DOS_21Handler,CB_IRET_STI);
+	CALLBACK_Setup(call_21,DOS_21Handler,CB_IRET_STI,"DOS Int 21");
 	RealSetVec(0x21,CALLBACK_RealPointer(call_21));
 
 	call_25=CALLBACK_Allocate();
-	CALLBACK_Setup(call_25,DOS_25Handler,CB_RETF);
+	CALLBACK_Setup(call_25,DOS_25Handler,CB_RETF,"DOS Int 25");
 	RealSetVec(0x25,CALLBACK_RealPointer(call_25));
 	
 	call_26=CALLBACK_Allocate();
-	CALLBACK_Setup(call_26,DOS_26Handler,CB_RETF);
+	CALLBACK_Setup(call_26,DOS_26Handler,CB_RETF,"DOS Int 26");
 	RealSetVec(0x26,CALLBACK_RealPointer(call_26));
 	
 	call_27=CALLBACK_Allocate();
-	CALLBACK_Setup(call_27,DOS_27Handler,CB_IRET);
+	CALLBACK_Setup(call_27,DOS_27Handler,CB_IRET,"DOS Int 27");
 	RealSetVec(0x27,CALLBACK_RealPointer(call_27));
 
-    call_28=CALLBACK_Allocate();
-    CALLBACK_Setup(call_28,DOS_28Handler,CB_IRET);
-    RealSetVec(0x28,CALLBACK_RealPointer(call_28));
+	call_28=CALLBACK_Allocate();
+	CALLBACK_Setup(call_28,DOS_28Handler,CB_IRET,"DOS Int 28");
+	RealSetVec(0x28,CALLBACK_RealPointer(call_28));
 
-    call_29=CALLBACK_Allocate();
-    CALLBACK_Setup(call_29,DOS_29Handler,CB_IRET);
-    RealSetVec(0x29,CALLBACK_RealPointer(call_29));
+	call_29=CALLBACK_Allocate();
+	CALLBACK_Setup(call_29,DOS_29Handler,CB_IRET,"CON Output Int 29");
+	RealSetVec(0x29,CALLBACK_RealPointer(call_29));
 
 	DOS_SetupFiles();								/* Setup system File tables */
 	DOS_SetupDevices();							/* Setup dos devices */
@@ -993,8 +1019,5 @@ void DOS_Init(Section* sec) {
 	/* shutdown function */
 	sec->AddDestroyFunction(&DOS_ShutDown);	
 
-	/* case map routine INT 0x21 0x38 */
-	call_casemap = CALLBACK_Allocate();
-    CALLBACK_Setup(call_casemap,DOS_CaseMapFunc,CB_RETF);
 }
 
