@@ -9,13 +9,15 @@
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU Library General Public License for more details.
+ *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
+/* $Id: dos_misc.cpp,v 1.12 2004/08/04 09:12:53 qbix79 Exp $ */
+
 #include "dosbox.h"
 #include "callback.h"
 #include "mem.h"
@@ -56,11 +58,56 @@ static Bitu INT2A_Handler(void) {
 
 static bool DOS_MultiplexFunctions(void) {
 	switch (reg_ax) {
+	case 0x1607:
+		if (reg_bx == 0x15) {
+			switch (reg_cx) {
+				case 0x0000:		// query instance
+					reg_cx = 0x0001;
+					reg_dx = 0x50;		// dos driver segment
+					SegSet16(es,0x50);	// patch table seg
+					reg_bx = 0x60;		// patch table ofs
+					return true;
+				case 0x0001:		// set patches
+					reg_ax = 0xb97c;
+					reg_bx = (reg_dx & 0x16);
+					reg_dx = 0xa2ab;
+					return true;
+				case 0x0003:		// get size of data struc
+					if (reg_dx==0x0001) {
+						// CDS size requested
+						reg_ax = 0xb97c;
+						reg_dx = 0xa2ab;
+						reg_cx = 0x000e;	// size
+					}
+					return true;
+				case 0x0004:		// instanced data
+					reg_dx = 0;		// none
+					return true;
+				case 0x0005:		// get device driver size
+					reg_ax = 0;
+					reg_dx = 0;
+					return true;
+				default:
+					return false;
+			}
+		}
+		else if (reg_bx == 0x18) return true;	// idle callout
+		else return false;
 	case 0x1680:	/*  RELEASE CURRENT VIRTUAL MACHINE TIME-SLICE */
 		//TODO Maybe do some idling but could screw up other systems :)
 		reg_al=0;	
 		return true;
+	case 0x1689:	/*  Kernel IDLE CALL */
 	case 0x168f:	/*  Close awareness crap */
+	   /* Removing warning */
+		return true;
+	case 0x4a01:	/* Query free hma space */
+	case 0x4a02:	/* ALLOCATE HMA SPACE */
+		LOG(LOG_DOSMISC,LOG_WARN)("INT 2f:4a HMA. DOSBox reports none available.");
+		reg_bx=0;	//number of bytes available in HMA or amount succesfully allocated
+		//ESDI=ffff:ffff Location of HMA/Allocated memory
+		SegSet16(es,0xffff);
+		reg_di=0xffff;
 		return true;
 	}
 
@@ -71,12 +118,12 @@ void DOS_SetupMisc(void) {
 	/* Setup the dos multiplex interrupt */
 	first_multiplex=0;
 	call_int2f=CALLBACK_Allocate();
-	CALLBACK_Setup(call_int2f,&INT2F_Handler,CB_IRET);
+	CALLBACK_Setup(call_int2f,&INT2F_Handler,CB_IRET,"DOS Int 2f");
 	RealSetVec(0x2f,CALLBACK_RealPointer(call_int2f));
 	DOS_AddMultiplexHandler(DOS_MultiplexFunctions);
 	/* Setup the dos network interrupt */
 	call_int2a=CALLBACK_Allocate();
-	CALLBACK_Setup(call_int2a,&INT2A_Handler,CB_IRET);
+	CALLBACK_Setup(call_int2a,&INT2A_Handler,CB_IRET,"DOS Int 2a");
 	RealSetVec(0x2A,CALLBACK_RealPointer(call_int2a));
 };
 
