@@ -10,30 +10,30 @@
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU Library General Public License for more details.
+ *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-/* $Id: drive_cache.cpp,v 1.32 2004/01/12 20:25:05 finsterr Exp $ */
+/* $Id: drive_cache.cpp,v 1.40 2004/11/13 12:08:43 qbix79 Exp $ */
 
 #include "drives.h"
 #include "dos_inc.h"
 #include "dirent.h"
 #include "support.h"
 
-#if defined (WIN32)   /* Win 32 */
-#define WIN32_LEAN_AND_MEAN        // Exclude rarely-used stuff from 
-#include <windows.h>
-#endif
-
 // STL stuff
 #include <vector>
 #include <iterator>
 #include <algorithm>
 
+#if defined (WIN32)   /* Win 32 */
+#define WIN32_LEAN_AND_MEAN        // Exclude rarely-used stuff from 
+#include <windows.h>
+#endif
+
 int fileInfoCounter = 0;
 
 bool SortByName(DOS_Drive_Cache::CFileInfo* const &a, DOS_Drive_Cache::CFileInfo* const &b) 
@@ -62,23 +62,27 @@ bool SortByDirNameRev(DOS_Drive_Cache::CFileInfo* const &a, DOS_Drive_Cache::CFi
 
 DOS_Drive_Cache::DOS_Drive_Cache(void)
 {
-	dirBase		= new CFileInfo;
-	save_dir	= 0;
-	srchNr		= 0;
-	label[0]	= 0;
+	dirBase			= new CFileInfo;
+	save_dir		= 0;
+	srchNr			= 0;
+	label[0]		= 0;
+	nextFreeFindFirst	= 0;
 	for (Bit32u i=0; i<MAX_OPENDIRS; i++) { dirSearch[i] = 0; free[i] = true; dirFindFirst[i] = 0; };
 	SetDirSort(DIRALPHABETICAL);
+	updatelabel = true;
 };
 
 DOS_Drive_Cache::DOS_Drive_Cache(const char* path)
 {
-	dirBase		= new CFileInfo;
-	save_dir	= 0;
-	srchNr		= 0;
-	label[0]	= 0;
+	dirBase			= new CFileInfo;
+	save_dir		= 0;
+	srchNr			= 0;
+	label[0]		= 0;
+	nextFreeFindFirst	= 0;
 	for (Bit32u i=0; i<MAX_OPENDIRS; i++) { dirSearch[i] = 0; free[i] = true; dirFindFirst[i] = 0; };
 	SetDirSort(DIRALPHABETICAL);
 	SetBaseDir(path);
+	updatelabel = true;
 };
 
 DOS_Drive_Cache::~DOS_Drive_Cache(void)
@@ -90,6 +94,7 @@ DOS_Drive_Cache::~DOS_Drive_Cache(void)
 void DOS_Drive_Cache::Clear(void)
 {
 	delete dirBase; dirBase = 0;
+	nextFreeFindFirst	= 0;
 	for (Bit32u i=0; i<MAX_OPENDIRS; i++) dirSearch[i] = 0;
 };
 
@@ -104,8 +109,14 @@ void DOS_Drive_Cache::EmptyCache(void)
 	SetBaseDir(basePath);
 };
 
-void DOS_Drive_Cache::SetLabel(const char* vname)
+void DOS_Drive_Cache::SetLabel(const char* vname,bool allowupdate)
 {
+/* allowupdate defaults to true. if mount sets a label then allowupdate is 
+ * false and will this function return at once after the first call.
+ * The label will be set at the first call. */
+
+	if(!this->updatelabel) return;
+	this->updatelabel = allowupdate;
 	Bitu togo		= 8;
 	Bitu vnamePos	= 0;
 	Bitu labelPos	= 0;
@@ -113,8 +124,7 @@ void DOS_Drive_Cache::SetLabel(const char* vname)
 	while (togo>0) {
 		if (vname[vnamePos]==0) break;
 		if (!point && (vname[vnamePos]=='.')) { togo=4; point=true; }
-		label[labelPos]	= vname[vnamePos];
-		label[labelPos] = *upcase(&label[labelPos]);
+		label[labelPos] = toupper(vname[vnamePos]);
 		labelPos++; vnamePos++;
 		togo--;
 		if ((togo==0) && !point) { 
@@ -123,7 +133,10 @@ void DOS_Drive_Cache::SetLabel(const char* vname)
 		}
 	};
 	label[labelPos]=0;
-//	LOG(LOG_ALL,LOG_ERROR)("CACHE: Set volume label to %s",label);
+	//Remove trailing dot.
+	if((labelPos > 0) && (label[labelPos-1] == '.'))
+		label[labelPos-1]=0;
+	LOG(LOG_DOSMISC,LOG_NORMAL)("DIRCACHE: Set volume label to %s",label);
 };
 
 Bit16u DOS_Drive_Cache::GetFreeID(CFileInfo* dir)
@@ -143,10 +156,13 @@ void DOS_Drive_Cache::SetBaseDir(const char* baseDir)
 	};
 	// Get Volume Label
 #if defined (WIN32)
-	char label[256];
+	char labellocal[256]={ 0 };
 	char drive[4] = "C:\\";
 	drive[0] = basePath[0];
-	if (GetVolumeInformation(drive,label,256,NULL,NULL,NULL,NULL,0)) SetLabel(label);
+	if (GetVolumeInformation(drive,labellocal,256,NULL,NULL,NULL,NULL,0)) {
+		/* Set label and allow being updated */
+		SetLabel(labellocal,true);
+	}
 #endif
 };
 
@@ -174,6 +190,17 @@ char* DOS_Drive_Cache::GetExpandName(const char* path)
 		GetLongName(dirInfo, dir);
 		strcat(work,dir);
 	}
+
+	if(work && *work) {
+		size_t len = strlen(work);
+#if defined (WIN32) 
+		if((work[len-1] == CROSS_FILESPLIT ) && (len >= 2) && (work[len-2] != ':')) {
+#else
+		if((len > 1) && (work[len-1] == CROSS_FILESPLIT )) {
+#endif       
+			work[len-1] = 0; // Remove trailing slashes except when in root
+		}
+	}
 	return work;
 };
 
@@ -431,8 +458,6 @@ void DOS_Drive_Cache::CreateShortName(CFileInfo* curDir, CFileInfo* info)
 		// Copy number
 		strcat(info->shortname,"~");
 		strcat(info->shortname,buffer);
-		// Create compare Count
-//		info->compareCount = tocopy;
 		// Add (and cut) Extension, if available
 		if (pos) {
 			// Step to last extension...
@@ -635,35 +660,29 @@ bool DOS_Drive_Cache::SetResult(CFileInfo* dir, char* &result, Bit16u entryNr)
 };
 
 // FindFirst / FindNext
-bool DOS_Drive_Cache::FindFirst(char* path, Bitu dtaAddress, Bitu& id)
+bool DOS_Drive_Cache::FindFirst(char* path, Bitu& id)
 {
 	Bit16u	dirID;
-	Bitu	dirFindFirstID = 0xffff;
+	Bitu	dirFindFirstID = this->nextFreeFindFirst++; //increase it for the next search
 
 	// Cache directory in 
 	if (!OpenDir(path,dirID)) return false;
-	// Seacrh if dta was already used before
-	for (Bitu n=0; n<MAX_OPENDIRS; n++) {
-		if (dirFindFirst[n]) {
-			if (dirFindFirst[n]->compareCount == dtaAddress) {
-				// Reuse old dta
-				dirFindFirstID = n; break;
-			}		
-		} else if (dirFindFirstID==0xffff) {
-			dirFindFirstID = n;
-		}	
-	} 
-	if (dirFindFirstID==0xffff) {
+
+	if (dirFindFirstID == MAX_OPENDIRS) {
 		// no free slot found...
-		LOG(LOG_MISC,LOG_ERROR)("DIRCACHE: FindFirst/Next failure : All slots full.");
-		// always use first then
+		LOG(LOG_MISC,LOG_ERROR)("DIRCACHE: FindFirst/Next: All slots full. Resetting");
+		// Clear the internal list then.
 		dirFindFirstID = 0;
+		this->nextFreeFindFirst = 1; //the next free one after this search
+		for(Bitu n=0; n<MAX_OPENDIRS;n++) {	
+	     		// Clear and reuse slot
+			delete dirFindFirst[n];
+			dirFindFirst[n]=0;
+		}
+	   
 	}		
-	// Clear and reuse slot
-	delete dirFindFirst[dirFindFirstID];
 	dirFindFirst[dirFindFirstID] = new CFileInfo();
 	dirFindFirst[dirFindFirstID]-> nextEntry	= 0;
-	dirFindFirst[dirFindFirstID]-> compareCount	= dtaAddress;	
 
 	// Copy entries to use with FindNext
 	for (Bitu i=0; i<dirSearch[dirID]->fileList.size(); i++) {
