@@ -9,14 +9,14 @@
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU Library General Public License for more details.
+ *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-/* $Id: ems.cpp,v 1.31 2004/02/02 20:23:54 qbix79 Exp $ */
+/* $Id: ems.cpp,v 1.36 2004/10/23 15:15:06 qbix79 Exp $ */
 
 #include <string.h>
 #include <stdlib.h>
@@ -65,7 +65,7 @@
 
 class device_EMM : public DOS_Device {
 public:
-	device_EMM(){name="EMMXXXX0";}
+	device_EMM(){SetName("EMMXXXX0");}
 	bool Read(Bit8u * data,Bit16u * size) { return false;}
 	bool Write(Bit8u * data,Bit16u * size){ 
 		LOG(LOG_IOCTL,LOG_NORMAL)("EMS:Write to device");	
@@ -150,9 +150,20 @@ static Bit8u EMM_MapPage(Bitu phys_page,Bit16u handle,Bit16u log_page) {
 //	LOG_MSG("EMS MapPage handle %d phys %d log %d",handle,phys_page,log_page);
 	/* Check for too high physical page */
 	if (phys_page>=EMM_MAX_PHYS) return EMM_ILL_PHYS;
+
+	/* unmapping doesn't need valid handle (as handle isn't used) */
+	if (log_page==NULL_PAGE) {
+		/* Unmapping */
+		emm_mappings[phys_page].handle=NULL_HANDLE;
+		emm_mappings[phys_page].page=NULL_PAGE;
+		for (Bitu i=0;i<4;i++) 
+			PAGING_MapPage(EMM_PAGEFRAME4K+phys_page*4+i,EMM_PAGEFRAME4K+phys_page*4+i);
+		PAGING_ClearTLB();
+		return EMM_NO_ERROR;
+	}
 	/* Check for valid handle */
 	if (!ValidHandle(handle)) return EMM_INVALID_HANDLE;
-	/* Check to do unmapping or mappning */
+	
 	if (log_page<emm_handles[handle].pages) {
 		/* Mapping it is */
 		emm_mappings[phys_page].handle=handle;
@@ -165,15 +176,7 @@ static Bit8u EMM_MapPage(Bitu phys_page,Bit16u handle,Bit16u log_page) {
 		}
 		PAGING_ClearTLB();
 		return EMM_NO_ERROR;
-	} else if (log_page==NULL_PAGE) {
-		/* Unmapping it is */
-		emm_mappings[phys_page].handle=NULL_HANDLE;
-		emm_mappings[phys_page].page=NULL_PAGE;
-		for (Bitu i=0;i<4;i++) 
-			PAGING_MapPage(EMM_PAGEFRAME4K+phys_page*4+i,EMM_PAGEFRAME4K+phys_page*4+i);
-		PAGING_ClearTLB();
-		return EMM_NO_ERROR;
-	} else {
+	} else  {
 		/* Illegal logical page it is */
 		return EMM_LOG_OUT_RANGE;
 	}
@@ -571,6 +574,14 @@ static Bitu INT67_Handler(void) {
 		reg_cx = EMM_MAX_PHYS;
 		reg_ah = EMM_NO_ERROR;
 		break;
+	case 0x5A:              /* Allocate standard/raw Pages */
+		if (reg_al==0x00) {
+			reg_ah=EMM_AllocateMemory(reg_bx,reg_dx);
+		} else {
+			LOG(LOG_MISC,LOG_ERROR)("EMS:Call 5A subfct %2X not supported",reg_al);
+			reg_ah=EMM_FUNC_NOSUP;
+		};
+		break;
 	case 0xDE:		/* VCPI Functions */
 		LOG(LOG_MISC,LOG_ERROR)("EMS:VCPI Call %2X not supported",reg_al);
 		reg_ah=EMM_FUNC_NOSUP;
@@ -599,14 +610,14 @@ static Bitu INT4B_Handler() {
 void EMS_Init(Section* sec) {
 	/* Virtual DMA interrupt callback */
 	call_vdma=CALLBACK_Allocate();
-	CALLBACK_Setup(call_vdma,&INT4B_Handler,CB_IRET);
+	CALLBACK_Setup(call_vdma,&INT4B_Handler,CB_IRET,"Int 4b vdma");
 	RealSetVec(0x4b,CALLBACK_RealPointer(call_vdma));
 
 	Section_prop * section=static_cast<Section_prop *>(sec);
 	if (!section->Get_bool("ems")) return;
 	BIOS_ZeroExtendedSize();
 	call_int67=CALLBACK_Allocate();	
-	CALLBACK_Setup(call_int67,&INT67_Handler,CB_IRET);
+	CALLBACK_Setup(call_int67,&INT67_Handler,CB_IRET,"Int 67 ems");
 /* Register the ems device */
 	DOS_Device * newdev = new device_EMM();
 	DOS_AddDevice(newdev);
