@@ -9,28 +9,37 @@
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU Library General Public License for more details.
+ *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-/* $Id: mouse.cpp,v 1.31 2004/01/31 22:42:49 harekiet Exp $ */
+/* $Id: mouse.cpp,v 1.42 2004/08/04 09:12:56 qbix79 Exp $ */
 
 #include <string.h>
+#include <math.h>
+
+
 #include "dosbox.h"
 #include "callback.h"
 #include "mem.h"
 #include "regs.h"
+#include "cpu.h"
 #include "mouse.h"
 #include "pic.h"
 #include "inout.h"
 #include "int10.h"
 #include "bios.h"
 
-static Bitu call_int33,call_int74;
 
+static Bitu call_int33,call_int74;
+static Bit16u ps2cbseg,ps2cbofs;
+static bool useps2callback;
+static Bit16u call_ps2;
+static RealPt ps2_callback;
+static Bit16s oldmouseX, oldmouseY;
 // forward
 void WriteMouseIntVector(void);
 
@@ -100,6 +109,8 @@ static struct {
 	float	mickeysPerPixel_y;
 	float	pixelPerMickey_x;
 	float	pixelPerMickey_y;
+	float	senv_x;
+	float	senv_y;
 	Bit16u  updateRegion_x[2];
 	Bit16u  updateRegion_y[2];
 	Bit16u  page;
@@ -110,6 +121,56 @@ static struct {
 	Bit16s oldshown;
 } mouse;
 
+void Mouse_SetPS2State(bool use) {
+	if ((SegValue(es)!=0) && (reg_bx!=0)) useps2callback = use;
+	else useps2callback = false;
+	Mouse_AutoLock(useps2callback);
+}
+
+void Mouse_ChangePS2Callback(Bit16u pseg, Bit16u pofs) {
+	if ((pseg==0) && (pofs==0)) useps2callback = false;
+	else useps2callback = true;
+	ps2cbseg = pseg;
+	ps2cbofs = pofs;
+	Mouse_AutoLock(useps2callback);
+}
+
+void DoPS2Callback(Bit16u data, Bit16s mouseX, Bit16s mouseY) {
+	if (useps2callback) {
+		Bit16u mdat = (data & 0x03) | 0x08;
+		Bit16s xdiff = mouseX-oldmouseX;
+		Bit16s ydiff = oldmouseY-mouseY;
+		oldmouseX = mouseX;
+		oldmouseY = mouseY;
+		if ((xdiff>0xff) || (xdiff<-0xff)) mdat |= 0x40;		// x overflow
+		if ((ydiff>0xff) || (ydiff<-0xff)) mdat |= 0x80;		// y overflow
+		xdiff %= 256;
+		ydiff %= 256;
+		if (xdiff<0) {
+			xdiff = (0x100+xdiff);
+			mdat |= 0x10;
+		}
+		if (ydiff<0) {
+			ydiff = (0x100+ydiff);
+			mdat |= 0x20;
+		}
+		CPU_Push16((Bit16u)mdat); 
+		CPU_Push16((Bit16u)(xdiff % 256)); 
+		CPU_Push16((Bit16u)(ydiff % 256)); 
+		CPU_Push16((Bit16u)0); 
+		CPU_Push16(RealSeg(ps2_callback));
+		CPU_Push16(RealOff(ps2_callback));
+		SegSet16(cs, ps2cbseg);
+		reg_ip = ps2cbofs;
+	}
+}
+
+Bitu PS2_Handler(void) {
+	reg_sp += 8;	// remove the 4 words
+	return CBRET_NONE;
+}
+
+
 #define X_MICKEY 8
 #define Y_MICKEY 8
 
@@ -283,7 +344,7 @@ void DrawCursor() {
 	// Get Clipping ranges
 
 	// In Textmode ?
-	if (CurMode->type<=M_TEXT16) {
+	if (CurMode->type==M_TEXT) {
 		DrawCursorText();
 		return;
 	}
@@ -344,19 +405,25 @@ void DrawCursor() {
 }
 
 void Mouse_CursorMoved(float x,float y) {
-
 	float dx = x * mouse.pixelPerMickey_x;
 	float dy = y * mouse.pixelPerMickey_y;
+	if((fabs(x) > 1.0) || (mouse.senv_y < 1.0)) dx *= mouse.senv_x;
+
+	if((fabs(y) > 1.0) || (mouse.senv_y < 1.0)) dy *= mouse.senv_y;
+
 
 	mouse.mickey_x += dx;
 	mouse.mickey_y += dy;
-
 	mouse.x += dx;
-	if (mouse.x>mouse.max_x) mouse.x=mouse.max_x;
-	if (mouse.x<mouse.min_x) mouse.x=mouse.min_x;
 	mouse.y += dy;
-	if (mouse.y>mouse.max_y) mouse.y=mouse.max_y;
-	if (mouse.y<mouse.min_y) mouse.y=mouse.min_y;
+	/* ignore constraints if using PS2 mouse callback in the bios */
+
+	if (!useps2callback) {		
+		if (mouse.x > mouse.max_x) mouse.x = mouse.max_x;
+		if (mouse.x < mouse.min_x) mouse.x = mouse.min_x;
+		if (mouse.y > mouse.max_y) mouse.y = mouse.max_y;
+		if (mouse.y < mouse.min_y) mouse.y = mouse.min_y;
+	}
 	Mouse_AddEvent(MOUSE_MOVED);
 	DrawCursor();
 }
@@ -407,8 +474,8 @@ void Mouse_ButtonReleased(Bit8u button) {
 	mouse.last_released_y[button]=POS_Y;
 }
 
-static void SetMickeyPixelRate(Bit16s px, Bit16s py)
-{
+static void SetMickeyPixelRate(Bit16s px, Bit16s py){
+
 	if ((px!=0) && (py!=0)) {
 		mouse.mickeysPerPixel_x	 = (float)px/X_MICKEY;
 		mouse.mickeysPerPixel_y  = (float)py/Y_MICKEY;
@@ -416,6 +483,15 @@ static void SetMickeyPixelRate(Bit16s px, Bit16s py)
 		mouse.pixelPerMickey_y 	 = Y_MICKEY/(float)py;	
 	}
 };
+static void SetSensitivity(Bit16s px, Bit16s py){
+	if ((px!=0) && (py!=0)) {
+		px--;  //Inspired by cutemouse 
+		py--;  //Although their cursor update routine is far more complex then ours
+		mouse.senv_x=(static_cast<float>(px)*px)/3600.0 +1.0/3.0;
+		mouse.senv_y=(static_cast<float>(py)*py)/3600.0 +1.0/3.0;
+     }
+};
+
 
 static void mouse_reset_hardware(void){
 	PIC_SetIRQMask(MOUSE_IRQ,false);
@@ -425,13 +501,11 @@ void Mouse_NewVideoMode(void)
 { //Does way to much. Many of this stuff should be moved to mouse_reset one day
 	WriteMouseIntVector();
    
-//	real_writed(0,(0x74<<2),CALLBACK_RealPointer(call_int74));
 	if(MOUSE_IRQ > 7) {
 		real_writed(0,((0x70+MOUSE_IRQ-8)<<2),CALLBACK_RealPointer(call_int74));
 	} else {
 		real_writed(0,((0x8+MOUSE_IRQ)<<2),CALLBACK_RealPointer(call_int74));
 	}
-
 	mouse.shown=-1;
 	/* Get the correct resolution from the current video mode */
 	Bitu mode=mem_readb(BIOS_VIDEO_MODE);
@@ -444,6 +518,9 @@ void Mouse_NewVideoMode(void)
 	case 0x05:
 	case 0x06:
 	case 0x07:
+	case 0x08:
+	case 0x09:
+	case 0x0a:
 	case 0x0d:
 	case 0x0e:
 	case 0x13:
@@ -462,15 +539,15 @@ void Mouse_NewVideoMode(void)
 		LOG(LOG_MOUSE,LOG_ERROR)("Unhandled videomode %X on reset",mode);
 		break;
 	} 
-	mouse.max_x=639;
-	mouse.min_x=0;
-	mouse.min_y=0;
+	mouse.max_x = 639;
+	mouse.min_x = 0;
+	mouse.min_y = 0;
 	// Dont set max coordinates here. it is done by SetResolution!
-	mouse.x=0;				// civ wont work otherwise
-	mouse.y=static_cast<float>(mouse.max_y/2);
-	mouse.events=0;
-	mouse.mickey_x=0;
-	mouse.mickey_y=0;
+	mouse.x = static_cast<float>(mouse.max_x / 2);
+	mouse.y = static_cast<float>(mouse.max_y / 2);
+	mouse.events = 0;
+	mouse.mickey_x = 0;
+	mouse.mickey_y = 0;
 
 	mouse.hotx		 = 0;
 	mouse.hoty		 = 0;
@@ -491,6 +568,12 @@ void Mouse_NewVideoMode(void)
 	mouse.oldshown=-1;
 
 	SetMickeyPixelRate(8,16);
+	oldmouseX = static_cast<Bit16s>(mouse.x);
+
+	oldmouseY = static_cast<Bit16s>(mouse.y);
+
+   
+
 }
 
 
@@ -502,6 +585,10 @@ static void mouse_reset(void) {
    	mouse.sub_mask=0;
 	mouse.sub_seg=0;
 	mouse.sub_ofs=0;
+	mouse.senv_x=1.0;
+
+	mouse.senv_y=1.0;
+
 
 	//Added this for cd-v19
 }
@@ -526,7 +613,7 @@ static Bitu INT33_Handler(void) {
 		break;
 	case 0x02:	/* Hide Mouse */
 		{
-			if (CurMode->type>M_TEXT16)		RestoreCursorBackground();
+			if (CurMode->type!=M_TEXT)		RestoreCursorBackground();
 			else							RestoreCursorBackgroundText();
 			mouse.shown--;
 		}
@@ -565,10 +652,11 @@ static Bitu INT33_Handler(void) {
 		}
 	case 0x07:	/* Define horizontal cursor range */
 		{	//lemmings set 1-640 and wants that. iron seeds set 0-640 but doesn't like 640
+			//Iron seed works if newvideo mode with mode 13 sets 0-639
+			//Larry 6 actually wants newvideo mode with mode 13 to set it to 0-319
 			Bits max,min;
 			if ((Bit16s)reg_cx<(Bit16s)reg_dx) { min=(Bit16s)reg_cx;max=(Bit16s)reg_dx;}
 			else { min=(Bit16s)reg_dx;max=(Bit16s)reg_cx;}
-			//if(max - min + 1 > 640) max = min + 640 - 1;
 			mouse.min_x=min;
 			mouse.max_x=max;
 			LOG(LOG_MOUSE,LOG_NORMAL)("Define Hortizontal range min:%d max:%d",min,max);
@@ -581,7 +669,6 @@ static Bitu INT33_Handler(void) {
 			Bits max,min;
 			if ((Bit16s)reg_cx<(Bit16s)reg_dx) { min=(Bit16s)reg_cx;max=(Bit16s)reg_dx;}
 			else { min=(Bit16s)reg_dx;max=(Bit16s)reg_cx;}
-		//	if(static_cast<Bitu>(max - min + 1) > CurMode->sheight) max = min + CurMode->sheight - 1;
 			mouse.min_y=min;
 			mouse.max_y=max;
 			LOG(LOG_MOUSE,LOG_NORMAL)("Define Vertical range min:%d max:%d",min,max);
@@ -665,12 +752,19 @@ static Bitu INT33_Handler(void) {
 		}
 		break;
 	case 0x1a:	/* Set mouse sensitivity */
-		SetMickeyPixelRate(reg_bx,reg_cx);
+		SetSensitivity(reg_bx,reg_cx);
+
+		LOG(LOG_MOUSE,LOG_WARN)("Set sensitivity used with %d %d",reg_bx,reg_cx);
+
 		// ToDo : double mouse speed value
 		break;
 	case 0x1b:	/* Get mouse sensitivity */
-		reg_bx = Bit16s(X_MICKEY * mouse.mickeysPerPixel_x);
-		reg_cx = Bit16s(Y_MICKEY * mouse.mickeysPerPixel_y);
+		reg_bx = Bit16s((60.0* sqrt(mouse.senv_x- (1.0/3.0)) ) +1.0);
+
+		reg_cx = Bit16s((60.0* sqrt(mouse.senv_y- (1.0/3.0)) ) +1.0);
+
+		LOG(LOG_MOUSE,LOG_WARN)("Get sensitivity %d %d",reg_bx,reg_cx);
+
 		// ToDo : double mouse speed value
 		reg_dx = 64;
 		break;
@@ -759,6 +853,8 @@ static Bitu INT74_Handler(void) {
 			SegSet16(ds,oldds); SegSet16(es,oldes); SegSet16(ss,oldss); // Save segments
 
 		}
+		DoPS2Callback(mouse.event_queue[mouse.events].buttons, POS_X, POS_Y);
+
 	}
 	IO_Write(0xa0,0x20);
 	IO_Write(0x20,0x20);
@@ -780,7 +876,8 @@ void CreateMouseCallback(void)
 {
 	// Create callback
 	call_int33=CALLBACK_Allocate();
-	CALLBACK_Setup(call_int33,&INT33_Handler,CB_IRET);
+	CALLBACK_Setup(call_int33,&INT33_Handler,CB_IRET,"Mouse");
+
 	// Create a mouse vector with weird address 
 	// for strange mouse detection routines in Sim City & Wasteland
 	Bit16u ofs = call_int33<<4;
@@ -797,15 +894,17 @@ void MOUSE_Init(Section* sec) {
 	
 	// Callback 0x33
 	CreateMouseCallback();
-
 	call_int74=CALLBACK_Allocate();
-	CALLBACK_Setup(call_int74,&INT74_Handler,CB_IRET);
+	CALLBACK_Setup(call_int74,&INT74_Handler,CB_IRET,"int 74");
 	if(MOUSE_IRQ > 7) {
 		real_writed(0,((0x70+MOUSE_IRQ-8)<<2),CALLBACK_RealPointer(call_int74));
 	} else {
 		real_writed(0,((0x8+MOUSE_IRQ)<<2),CALLBACK_RealPointer(call_int74));
 	}
-
+	useps2callback = false;
+ 	call_ps2=CALLBACK_Allocate();
+	CALLBACK_Setup(call_ps2,&PS2_Handler,CB_IRET,"ps2 bios callback");
+	ps2_callback=CALLBACK_RealPointer(call_ps2);
 	memset(&mouse,0,sizeof(mouse));
 	mouse_reset_hardware();
 	mouse_reset();
